<?php

namespace Drp\NativeEventApp\Models\Concerns;

use Drp\NativeEventApp\Database\TableName;
use Drp\NativeEventApp\Models\Segment;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Relations\BelongsToMany;
use Illuminate\Support\Collection;

trait Segmentable
{
    public function segments(): BelongsToMany
    {
        return $this->belongsToMany(
            Segment::class,
            TableName::make('segments_' . str_replace(TableName::prefix(), '', $this->table))
        );
    }

    /**
     * @param mixed $segmentable
     */
    public function scopeViewableBy(Builder $query, $segmentable): Builder
    {
        return $query->whereHas('segments', function (Builder $subQuery) use ($segmentable): Builder {
            return $subQuery->whereNotIn(
                $this->segments()->getRelated()->getQualifiedKeyName(),
                $segmentable->segments->map->id
            );
        }, '=', 0);
    }

    public function scopeNotSegmented(Builder $query): Builder
    {
        return $query->whereDoesntHave('segments');
    }

    public function scopeHasAllSegments(Builder $query, Collection $segments): Builder
    {
        return $query->whereHas('segments', function (Builder $subQuery) use ($segments): Builder {
            return $subQuery->whereIn(
                $this->segments()->getRelated()->getQualifiedKeyName(),
                $segments->map->id
            );
        }, '=', $segments->count());
    }

    public function scopeHasSegment(Builder $query, Segment $segment): Builder
    {
        return $query->hasAllSegments(collect([$segment]));
    }

    public function joinSegments(Collection $segments): void
    {
        $segmentIds = $segments->keyBy->id->map->id;
        $attachedSegmentIds = $this->segments()
            ->whereIn(TableName::make('segments') . '.id', $segmentIds)
            ->get()
            ->map(static function ($segmentRegistration) {
                return (int) $segmentRegistration->id;
            });

        $this->segments()->attach($segmentIds->except($attachedSegmentIds));
    }

    public function syncSegments(Collection $segments): void
    {
        $this->segments()->sync($segments->pluck('id')->toArray());
    }

    public function hasAllSegments(Collection $segments): bool
    {
        $segmentIds = $segments->map->id->unique();
        $segmentRegistrations = $this->segments()
            ->whereIn(TableName::make('segments') . '.id', $segmentIds)
            ->count();

        return $segmentIds->count() === $segmentRegistrations;
    }

    public function hasAtLeastOneSegment(Collection $segments): bool
    {
        return $this->segments()
            ->whereIn(TableName::make('segments') . '.id', $segments->map->id)
            ->count() > 0;
    }
}
