<?php

namespace Drp\NativeEventApp\Test\Unit\Models\Concerns;

use Carbon\Carbon;
use Drp\NativeEventApp\Models\Registration;
use Drp\NativeEventApp\Models\Segment;
use Drp\NativeEventApp\Test\TestCase;

class SegmentableTest extends TestCase
{
    /** @var Registration */
    protected $registration;

    protected function setUp(): void
    {
        parent::setUp();

        Carbon::setTestNow('30-06-2020 13:00');

        $this->registration = factory(Registration::class)->create();
    }

    /** @test */
    public function it_can_join_segments_onto_a_segmentable_model(): void
    {
        [$segmentA, $segmentB] = factory(Segment::class)->times(2)->create();

        $this->registration->joinSegments(collect([$segmentA, $segmentB]));

        $segments = $this->registration->segments()->get();

        $this->assertCount(2, $segments);
        $this->assertEquals($segmentA->id, $segments->first()->id);
        $this->assertEquals($segmentB->id, $segments->last()->id);
    }

    /** @test */
    public function it_will_not_join_segments_onto_a_segmentable_model_that_already_exist(): void
    {
        [$segmentA, $segmentB] = factory(Segment::class)->times(2)->create();

        $this->registration->joinSegments(collect([$segmentA, $segmentB]));

        $segments = $this->registration->segments()->get();

        $this->assertCount(2, $segments);

        $this->registration->joinSegments(collect([$segmentA]));

        $this->assertCount(2, $segments);
    }

    /** @test */
    public function it_can_check_if_a_segmentable_model_has_all_segments(): void
    {
        [$segmentA, $segmentB] = factory(Segment::class)->times(2)->create();

        $this->registration->joinSegments(collect([$segmentA, $segmentB]));

        $this->assertTrue($this->registration->hasAllSegments(collect([$segmentA])));
    }

    /** @test */
    public function it_can_check_if_a_segmentable_model_has_at_least_one_segment(): void
    {
        [$segmentA, $segmentB] = factory(Segment::class)->times(2)->create();

        $this->registration->joinSegments(collect([$segmentA, $segmentB]));

        $this->assertTrue($this->registration->hasAtLeastOneSegment(collect([$segmentB])));
    }

    /** @test */
    public function it_can_sync_segments_onto_a_segmentable_model(): void
    {
        [$segmentA, $segmentB] = factory(Segment::class)->times(2)->create();
        [$segmentC, $segmentD] = factory(Segment::class)->times(2)->create();

        $this->registration->syncSegments(collect([$segmentA, $segmentB]));

        $segments = $this->registration->segments()->get();

        $this->assertCount(2, $segments);
        $this->assertEquals($segmentA->id, $segments->first()->id);
        $this->assertEquals($segmentB->id, $segments->last()->id);

        $this->registration->syncSegments(collect([$segmentC, $segmentD]));

        $segments = $this->registration->segments()->get();

        $this->assertCount(2, $segments);
        $this->assertEquals($segmentC->id, $segments->first()->id);
        $this->assertEquals($segmentD->id, $segments->last()->id);
    }
}
