<?php

namespace Drp\NativeEventApp\Test\Unit\Notifications;

use Carbon\Carbon;
use Drp\NativeEventApp\Database\TableName;
use Drp\NativeEventApp\Enums\NotificationStatus;
use Drp\NativeEventApp\Exceptions\NotificationNotScheduled;
use Drp\NativeEventApp\Models\Device;
use Drp\NativeEventApp\Models\Event;
use Drp\NativeEventApp\Models\Notification;
use Drp\NativeEventApp\Models\Registration;
use Drp\NativeEventApp\Models\Segment;
use Drp\NativeEventApp\Templates\DataProviders\DataProvider;
use Drp\NativeEventApp\Templates\DataProviders\DataProviderRegistry;
use Drp\NativeEventApp\Test\Support\InteractsWithEvents;
use Drp\NativeEventApp\Test\TestCase;
use Illuminate\Support\Facades\DB;

class NotificationTest extends TestCase
{
    use InteractsWithEvents;

    /** @test */
    public function it_can_make_a_new_notification(): void
    {
        $event = factory(Event::class)->create();
        $title = 'Test Title';
        $message = 'Test Messsage';
        Notification::make($event, $title, $message)->save();

        $this->assertEquals(1, Notification::count());

        $notification = Notification::find(1);

        $this->assertEquals($title, $notification->title);
        $this->assertEquals($message, $notification->message);
        $this->assertEquals(NotificationStatus::DRAFT, $notification->status);
        $this->assertNull($notification->send_at);
    }

    /** @test */
    public function it_can_ensure_the_model_is_persisted(): void
    {
        $event = factory(Event::class)->create();
        $notification = Notification::make($event, 'Test Title', 'Test Message')
            ->sendAt(now()->addHour());

        $this->assertNull($notification->id);

        $notification = $notification->ensureModelExists();

        $this->assertEquals(1, $notification->id);
    }

    /** @test */
    public function it_can_assign_a_segment_to_a_notification(): void
    {
        $event = factory(Event::class)->create();
        $segment = factory(Segment::class)->create();

        Notification::make($event, 'Test Title', 'Test Message')
            ->forSegment($segment);

        $this->assertEquals(1, Notification::count());

        $notification = Notification::find(1);

        $this->assertCount(1, $notification->segments);
        $this->assertEquals(
            $notification->id,
            $notification->segments()->withPivot('notification_id')->first()->pivot->notification_id
        );
        $this->assertEquals(
            $segment->id,
            $notification->segments()->withPivot('segment_id')->first()->pivot->segment_id
        );
        $this->assertEquals($segment->id, $notification->segments()->first()->id);
    }

    /** @test */
    public function it_can_assign_multiple_segments_to_a_notification(): void
    {
        $event = factory(Event::class)->create();
        $segments = factory(Segment::class)->times(3)->create();

        Notification::make($event, 'Test Title', 'Test Message')
            ->forSegments($segments);

        $this->assertEquals(1, Notification::count());

        $notification = Notification::find(1);

        $this->assertCount(3, $notification->segments);
        $this->assertEquals(
            $notification->id,
            $notification->segments()->withPivot('notification_id')->first()->pivot->notification_id
        );
        $this->assertEquals(
            $segments->first()->id,
            $notification->segments()->withPivot('segment_id')->first()->pivot->segment_id
        );
        $this->assertEquals($segments->first()->id, $notification->segments()->first()->id);
    }

    /** @test */
    public function it_can_assign_a_device_to_a_notification(): void
    {
        $event = factory(Event::class)->create();
        $device = factory(Device::class)->create();

        Notification::make($event, 'Test Title', 'Test Message')
            ->forDevice($device);

        $this->assertEquals(1, Notification::count());

        $notification = Notification::find(1);

        $this->assertCount(1, $notification->devices);
        $this->assertEquals(
            $notification->id,
            $notification->devices()->withPivot('notification_id')->first()->pivot->notification_id
        );
        $this->assertEquals(
            $device->id,
            $notification->devices()->withPivot('device_id')->first()->pivot->device_id
        );
        $this->assertEquals($device->id, $notification->devices()->first()->id);
    }

    /** @test */
    public function it_can_be_scheduled_to_send_at_the_specified_time(): void
    {
        $event = factory(Event::class)->create();
        Carbon::setTestNow('31-03-2020 13:00');

        $sendAt = now()->addHour();

        Notification::make($event, 'Test Title', 'Test Message')
            ->sendAt($sendAt)
            ->save();

        $this->assertEquals(1, Notification::count());

        $notification = Notification::find(1);

        $this->assertEquals($sendAt, $notification->send_at);
        $this->assertEquals(NotificationStatus::DRAFT, $notification->status);
    }

    /** @test */
    public function it_cannot_be_scheduled_to_send_in_the_past(): void
    {
        Carbon::setTestNow('31-03-2020 13:00');
        $event = factory(Event::class)->create();

        $sendAt = now()->subHour();

        try {
            Notification::make($event, 'Test Title', 'Test Message')
                ->sendAt($sendAt)
                ->save();
        } catch (\InvalidArgumentException $exception) {
            $this->assertEquals('The send at date must not be in the past', $exception->getMessage());
            $this->assertEquals(0, Notification::count());

            return;
        }

        $this->fail('InvalidArgumentException not thrown, date was set in the past');
    }

    /** @test */
    public function it_can_set_a_notification_to_published_and_save(): void
    {
        Carbon::setTestNow('31-03-2020 13:00');
        $event = factory(Event::class)->create();

        $sendAt = now()->addHour();

        Notification::make($event, 'Test Title', 'Test Message')
            ->sendAt($sendAt)
            ->publish();

        $this->assertEquals(1, Notification::count());

        $notification = Notification::find(1);

        $this->assertEquals($sendAt, $notification->send_at);
        $this->assertEquals(NotificationStatus::PUBLISHED, $notification->status);
    }

    /** @test */
    public function it_cannot_be_published_if_a_send_at_date_has_not_been_set(): void
    {
        Carbon::setTestNow('31-03-2020 13:00');
        $event = factory(Event::class)->create();

        try {
            Notification::make($event, 'Test Title', 'Test Message')
                ->publish();
        } catch (NotificationNotScheduled $exception) {
            $this->assertEquals('The send at date must be set before publishing', $exception->getMessage());
            $this->assertEquals(0, Notification::count());

            return;
        }

        $this->fail('NotificationNotScheduled not thrown and the send at date has not been set');
    }

    /** @test */
    public function it_can_unpublish_a_notification(): void
    {
        $notification = factory(Notification::class)->state('published')->create();

        $this->assertEquals(NotificationStatus::PUBLISHED, $notification->status);

        $notification = $notification->unpublish();

        $this->assertEquals(NotificationStatus::DRAFT, $notification->status);
    }

    /** @test */
    public function it_can_send_a_notification_now(): void
    {
        Carbon::setTestNow('31-03-2020 13:00');

        $event = factory(Event::class)->create();
        $device = factory(Device::class)->create();

        Notification::make($event, 'Test Title', 'Test Message')
            ->forDevice($device)
            ->sendNow();

        $this->assertEquals(1, Notification::count());

        $notification = Notification::find(1);

        $this->assertEquals(now()->addSecond(), $notification->send_at);
        $this->assertEquals(NotificationStatus::PUBLISHED, $notification->status);
    }

    /** @test */
    public function it_can_set_the_total_sent(): void
    {
        $notification = factory(Notification::class)->state('published')->create();

        $notification = $notification->totalSent(100);

        $this->assertEquals(100, $notification->total_sent);
    }

    /** @test */
    public function it_can_get_a_notification(): void
    {
        $event = factory(Event::class)->create();

        $notification = Notification::make($event, 'Title', 'Message')
            ->sendAt(now()->addHour())
            ->publish();

        $result = Notification::get($notification->id);

        $this->assertEquals($notification->id, $result->id);
        $this->assertEquals($notification->title, $result->title);
        $this->assertEquals(NotificationStatus::PUBLISHED, $result->status);
    }

    /** @test */
    public function it_can_get_all_notifications(): void
    {
        $eventA = factory(Event::class)->create();
        $eventB = factory(Event::class)->create();

        $notificationA = Notification::make($eventA, 'Title', 'A')
            ->sendAt(now()->addHour())
            ->publish();

        $notificationB = Notification::make($eventB, 'Title', 'B')
            ->sendAt(now()->addHours(2))
            ->publish();

        $results = Notification::get();

        $this->assertEquals(2, $results->count());
        $this->assertEquals($notificationA->id, $results->first()->id);
        $this->assertEquals($notificationA->message, $results->first()->message);
        $this->assertEquals(NotificationStatus::PUBLISHED, $results->first()->status);
        $this->assertEquals($notificationB->id, $results->last()->id);
        $this->assertEquals($notificationB->message, $results->last()->message);
        $this->assertEquals(NotificationStatus::PUBLISHED, $results->last()->status);
    }

    /** @test */
    public function it_can_get_all_unsent_notifications(): void
    {
        $event = factory(Event::class)->create();
        $notificationA = Notification::make($event, 'Title', 'A')
            ->sendAt(now()->addHour())
            ->publish();

        $notificationB = Notification::make($event, 'Title', 'B')
            ->sendAt(now()->addHours(2))
            ->publish();

        $results = Notification::getUnsent();

        $this->assertEquals(2, $results->count());
        $this->assertEquals($notificationA->id, $results->first()->id);
        $this->assertNull($results->first()->sent_at);
        $this->assertEquals($notificationB->id, $results->last()->id);
        $this->assertNull($results->last()->sent_at);
    }

    /** @test */
    public function it_can_get_all_unsent_notifications_with_no_segment(): void
    {
        $event = factory(Event::class)->create();
        $segment = factory(Segment::class)->create();

        $notificationA = Notification::make($event, 'Title', 'A')
            ->forSegment($segment)
            ->sendAt(now()->addHour())
            ->publish();

        $notificationB = Notification::make($event, 'Title', 'B')
            ->sendAt(now()->addHours(2))
            ->publish();

        $results = Notification::getUnsentWithoutSegment();

        $this->assertEquals(1, $results->count());
        $this->assertNotContains($notificationA, $results);
        $this->assertEquals($notificationB->id, $results->first()->id);
        $this->assertNull($results->first()->sent_at);
    }

    /** @test */
    public function it_can_get_all_sent_notifications(): void
    {
        Carbon::setTestNow('31-03-2020 13:00');
        $event = factory(Event::class)->create();

        $notificationA = Notification::make($event, 'Title', 'A')
            ->sendNow()
            ->publish();
        $notificationA->update(['sent_at' => now()]);

        $notificationB = Notification::make($event, 'Title', 'B')
            ->sendNow()
            ->publish();
        $notificationB->update(['sent_at' => now()]);

        $results = Notification::getSent();

        $this->assertEquals(2, $results->count());
        $this->assertEquals($notificationA->id, $results->first()->id);
        $this->assertEquals(now(), $results->first()->sent_at);
        $this->assertEquals($notificationB->id, $results->last()->id);
        $this->assertEquals(now(), $results->last()->sent_at);
    }

    /** @test */
    public function it_can_destroy_a_notification(): void
    {
        Carbon::setTestNow('31-03-2020 13:00');
        $event = factory(Event::class)->create();
        $segment = factory(Segment::class)->create();
        $device = factory(Device::class)->create();

        $notification = Notification::make($event, 'Title', 'A')
            ->forDevice($device)
            ->forSegment($segment)
            ->sendNow()
            ->publish();

        $this->assertEquals(1, Notification::count());
        $this->assertEquals(1, DB::table(TableName::make('segments_notifications'))->count());
        $this->assertEquals(1, DB::table(TableName::make('device_notifications'))->count());

        $result = $notification->delete();

        $this->assertTrue($result);
        $this->assertEquals(0, Notification::count());
        $this->assertEquals(0, DB::table(TableName::make('segments_notifications'))->count());
        $this->assertEquals(0, DB::table(TableName::make('device_notifications'))->count());
    }

    /** @test */
    public function it_can_get_notifications_by_a_segment(): void
    {
        $event = factory(Event::class)->create();
        $segmentA = factory(Segment::class)->create();
        $segmentB = factory(Segment::class)->create();

        Notification::make($event,'Title', 'A')
            ->forSegment($segmentA)
            ->sendAt(now()->addHour())
            ->publish();

        Notification::make($event, 'Title', 'B')
            ->forSegment($segmentA)
            ->sendAt(now()->addHours(2))
            ->publish();

        Notification::make($event, 'Title', 'C')
            ->forSegment($segmentB)
            ->sendAt(now()->addHours(2))
            ->publish();

        $this->assertEquals(3, DB::table(TableName::make('segments_notifications'))->count());

        $results = Notification::getBySegment($segmentA);

        $this->assertCount(2, $results);
        $first = $results->first();
        $last = $results->last();
        $this->assertEquals('A', $first->message);
        $this->assertEquals('B', $last->message);
    }

    /** @test */
    public function it_can_get_all_notifications_that_are_not_segmented(): void
    {
        $event = factory(Event::class)->create();
        $segment = factory(Segment::class)->create();

        Notification::make($event, 'Title', 'A')
            ->forSegment($segment)
            ->sendAt(now()->addHour())
            ->publish();

        Notification::make($event, 'Title', 'B')
            ->sendAt(now()->addHours(2))
            ->publish();

        Notification::make($event, 'Title', 'C')
            ->sendAt(now()->addHours(2))
            ->publish();

        $results = Notification::notSegmented()->get();

        $this->assertCount(2, $results);
        $this->assertEquals('B', $results->first()->message);
        $this->assertEquals('C', $results->last()->message);
    }

    /** @test */
    public function it_can_unschedule_a_notification_due_to_be_sent(): void
    {
        $event = factory(Event::class)->create();
        $notification = Notification::make($event, 'Title', 'A')
            ->sendAt(now()->addHours(2))
            ->publish();

        $this->assertNotNull($notification->send_at);
        $this->assertEquals(NotificationStatus::PUBLISHED, $notification->status);

        $notification = $notification->unschedule();

        $this->assertNull($notification->send_at);
        $this->assertEquals(NotificationStatus::PUBLISHED, $notification->status);
    }

    /** @test */
    public function it_can_get_the_sendable_notifications(): void
    {
        Carbon::setTestNow('02-04-2020 14:00');
        $event = factory(Event::class)->create();

        // Should send
        Notification::make($event, 'Title', 'A')
            ->sendAt(now())
            ->publish();

        $notificationB = Notification::make($event, 'Title', 'B')
            ->fill(['send_at' => now()->subHour()])
            ->publish();

        // Should not send
        Notification::make($event, 'Title', 'C')
            ->sendAt(now()->addHour())
            ->publish();

        $results = Notification::toBeSent()->get();

        $this->assertCount(3, Notification::getUnsent());
        $this->assertCount(2, $results);
        $this->assertEquals($notificationB->message, $results->last()->message);
    }

    /** @test */
    public function it_can_mark_a_notification_as_sent(): void
    {
        Carbon::setTestNow('06-04-2020 13:00');

        $notification = factory(Notification::class)->state('published')->create();

        $notification->markAsSent();

        $this->assertEquals(now(), $notification->sent_at);
    }

    /** @test */
    public function it_can_get_a_notifications_event(): void
    {
        Carbon::setTestNow('06-04-2020 13:00');

        $event = factory(Event::class)->create();

        $notification = Notification::make($event, 'Title', 'A')
            ->sendAt(now())
            ->publish();

        $this->assertEquals($event->id, $notification->event->id);
    }

    /** @test */
    public function it_can_render_personalised_content_in_the_title(): void
    {
        Carbon::setTestNow('06-04-2020 13:00');

        [$user, $event, $registration] = $this->registerFakeEventWithRegistration([
            'forename' => 'Dave',
        ]);

        $registry = app(DataProviderRegistry::class);
        $registry->bind('registration', static function (Registration $registration) {
            return new class(['forename' => $registration->forename]) extends DataProvider {
                /** @var string */
                public $forename;
            };
        });

        $title = 'Hello {registration.forename}, this is the title';
        $expected = 'Hello Dave, this is the title';

        $notification = Notification::make($event, $title, 'Test Message')
            ->sendAt(now()->addHour())
            ->publish();

        $this->assertEquals($expected, $notification->renderTitle($registration));
    }

    /** @test */
    public function it_can_render_personalised_content_in_the_message(): void
    {
        Carbon::setTestNow('06-04-2020 13:00');

        [$user, $event, $registration] = $this->registerFakeEventWithRegistration([
            'forename' => 'Dave',
        ]);

        $registry = app(DataProviderRegistry::class);
        $registry->bind('registration', static function (Registration $registration) {
            return new class(['forename' => $registration->forename]) extends DataProvider {
                /** @var string */
                public $forename;
            };
        });

        $message = 'Hello {registration.forename}, this is the message';
        $expected = 'Hello Dave, this is the message';

        $notification = Notification::make($event, 'Test Title', $message)
            ->sendAt(now()->addHour())
            ->publish();

        $this->assertEquals($expected, $notification->renderMessage($registration));
    }
}
